"""
Module to set up pvlib models for CCRenew projects and coordinate running
calculations.
"""

from __future__ import annotations

import numpy as np
import pandas as pd
from pvlib import (
    irradiance,
    location,
    pvsystem
)
from pvlib.modelchain import ModelChain
import sys

from ccrenew import (
    DaylightParams,
    DateLike,
    Numeric
)
from ccrenew import (
    cloud_data,
    data_determination as det
)

class ProjectModel(ModelChain):
    """An extension of the pvlib ModelChain class tailored to CCR sites.

    Args:
        project_name (str): The name of the project to model.
        azimuth (Numeric): Azimuth of the modules. North=0, East=90, South=180, West=270.
        axis_tilt (Numeric): For tracker sites -
            the angle that the axis of rotation makes with respect to horizontal.
        cross_axis_tilt (Numeric): For tracker sites on sloped surfaces -
            the angle that forms the slope of the ground perpendicular to the axis of rotation.
        ground_surface_type (str | None): Ground surface type to use for `ground_surface_albedo`.
            See `pvlib.irradiance.SURFACE_ALBEDOS` for valid values.
        ground_surface_albedo (float | None): Ground surface albedo to use for reflected irradiance.
            If `None` then `ground_surface_type` is used to look up a value.
            If `ground_surface_type` is also None, a default value of 0.25 is used.
        clearsky_model (str): Model used to calculate clearsky estimates of GHI, DNI, and/or DHI at the location.
        transposition_model (str): Irradiance model to use for transposing horizontal irradiance to POA irradiance.
            See: <a href="https://pvlib-python.readthedocs.io/en/stable/reference/generated/pvlib.irradiance.get_total_irradiance.html" target="_blank" rel="noopener noreferrer">Get Total Irradiance.</a>
        solar_position_method (str): Method to use for calculating solar position.
            See: <a href="https://pvlib-python.readthedocs.io/en/stable/reference/generated/pvlib.solarposition.get_solarposition.html#pvlib.solarposition.get_solarposition" target="_blank" rel="noopener noreferrer">Solar Position.</a>
        airmass_model (str): Model used for calculating airmass based on location.
            See: <a href="https://pvlib-python.readthedocs.io/en/stable/reference/generated/pvlib.location.Location.get_airmass.html" target="_blank" rel="noopener noreferrer">Get Airmass.</a>
        dc_model (str | None): Model used for DC power calculations. If `None`, model will be inferred from module parameters.
            See `pvsystem._DC_MODEL_PARAMS` for valid values.
        ac_model (str | None): Model used for AC power calculations. If `None`, model will be inferred from inverter parameters.
            Valid values are `sandia`, `adr`, or `pvwatts`.
        aoi_model (str | None): Model for calculating angle of incidence. If `None`, model will be inferred from module parameters.
            Valid values are `physical`, `ashrae`, `sapm`, `martin_ruiz`, or `no_loss`.
        spectral_model (str): Model for calculating spectral mismatch between test conditions and actual conditions in the field. If `None`, model will be inferred from module parameters.
            Valid values are `sapm`, `first_solar`, or `no_loss`.
        temperature_model (str): Model used to correct module efficiency based on cell temperature. If `None`, model will be inferred from parameters passed to the temperature model of the array.
            Valid values are `sapm`, `pvsyst`, `faiman`, `fuentes`, `noct_sam`.
        dc_ohmic_model (str): Model for calculating losses based on equivalent resistence.
            Valid values are `dc_ohms_from_percent` or `no_loss`
        losses_model (str): Model for calculating system losses.
            For PVWatts model losses see <a href="https://pvlib-python.readthedocs.io/en/stable/reference/pv_modeling/generated/pvlib.pvsystem.pvwatts_losses.html#pvlib.pvsystem.pvwatts_losses" target="_blank" rel="noopener noreferrer">Losses Model.</a>
            Valid values are `pvwatts` or `no_loss`
    """
    def __init__(self, project_name:str, azimuth:Numeric = 180, axis_tilt:Numeric = 0, cross_axis_tilt:Numeric = 0,
                 ground_surface_type:str|None = 'grass', ground_surface_albedo:float|None = None,
                 clearsky_model:str = 'ineichen', transposition_model:str = 'haydavies',
                 solar_position_method:str = 'nrel_numpy', airmass_model:str = 'kastenyoung1989',
                 dc_model:str|None = None, ac_model:str|None = None, aoi_model:str|None = 'physical',
                 spectral_model:str = 'no_loss', temperature_model:str|None = None,
                 dc_ohmic_model:str = 'no_loss', losses_model:str = 'no_loss'):

        # Get parameters for the project to pass to the pvlib constructors
        project_params = cloud_data._get_project_model_params(project_name)

        #choose tracking algorithm based upon temp_coefficient of the modules
        backtrack=True
        if project_params.temp_coef > -0.39:
            backtrack=False
  
        self.racking = project_params.racking
        if self.racking == 'Fixed':
            mount = pvsystem.FixedMount(project_params.tilt_gcr, azimuth)
        else:
            mount = pvsystem.SingleAxisTrackerMount(
                axis_tilt, azimuth, project_params.max_angle,
                backtrack, project_params.tilt_gcr, cross_axis_tilt)

        self.array = pvsystem.Array(mount, surface_type=ground_surface_type, name=project_name,
            module_parameters={'pdc0': project_params.mwdc*1000,
                            'gamma_pdc': project_params.temp_coef/100,
                            'dc_model': dc_model, 'ac_model': ac_model,
                            'aoi_model': aoi_model, 'spectral_model': spectral_model,
                            'dc_ohmic_model': dc_ohmic_model, 'losses_model': losses_model},
            temperature_model_parameters={'a': project_params.a_module,
                                          'b': project_params.b_module,
                                          'deltaT': project_params.delta_tcnd})

        self.pv_system = pvsystem.PVSystem([self.array], name=project_name,
            inverter_parameters={'pdc0': project_params.mwdc*1000})

        self.location = location.Location(
            latitude=project_params.lat, longitude=project_params.lon,
            tz=project_params.tz, altitude=project_params.elevation, name=project_name
        )

        super().__init__(self.pv_system, self.location,
                         clearsky_model, transposition_model,
                         solar_position_method, airmass_model,
                         dc_model, ac_model, aoi_model,
                         spectral_model, temperature_model,
                         dc_ohmic_model, losses_model, project_name)

        if isinstance(self.array.mount, pvsystem.FixedMount):
            self.racking = 'fixed'
        else:
            self.racking = 'tracker'


    def calculate_poa_from_ghi(self, ghi:pd.Series, model:str = 'Perez', shift:bool = False, **kwargs):
        # TODO: get model from system instead of supplying it here
        # `get_solarposition()` returns a df of `azimuth`, `apparent_elevation`, `elevation`, `apparent_zenith`, & `zenith`
        # apparent zenith & apparent elevation account for atmospheric refraction
        ghi = ghi.tz_localize(self.location.tz)
        if shift:
            ghi.index = ghi.index + pd.Timedelta('30min')
        solar_position = self.location.get_solarposition(ghi.index)
        
        # The erbs model takes the zenith un-corrected for refraction
        solar_zenith = solar_position['zenith']
        solar_azimuth = solar_position['azimuth']
        irradiation_components = irradiance.erbs(ghi, solar_zenith, ghi.index)
        dni = irradiation_components['dni']
        dhi = irradiation_components['dhi']

        poa = self.array.get_irradiance(solar_zenith, solar_azimuth, dni, ghi, dhi, model=model, **kwargs)

        if shift:
            poa.index = poa.index - pd.Timedelta('30min') # type: ignore
    
        poa = poa.tz_localize(None) # type: ignore

        return poa

    def run_bluesky(self, start_date:str|DateLike, end_date:str|DateLike,
                    resample:bool = True, fetch_ghi:bool = False) -> pd.DataFrame:
        print(f"Fetching solcast data for {self.name}")
        df_weather = cloud_data.get_sat_weather(self.name, start_date, end_date) # type: ignore
        sat_poa = self.calculate_poa_from_ghi(df_weather['sat_ghi'])
        df_solcast = df_weather.join(sat_poa['poa_global'])
        df_solcast = df_solcast.rename(columns={'poa_global': 'sat_poa'})
        df_solcast['poa'] = df_solcast['sat_poa']
        
        if fetch_ghi:
            try:
                print(f"Fetching 5 min GHI data for {self.name}")
                df_project_ghi = cloud_data.get_project_ghi(self.name, start_date, end_date) # type: ignore
                project_poa = self.calculate_poa_from_ghi(df_project_ghi['proj_ghi'])
                df_solcast['proj_ghi_poa'] = project_poa['poa_global']

                # Pass transposed GHI column through daylight zeroes filter to make sure the data is good
                # (the filter returns True where there are daylight zeroes so we need to invert it)
                proj_ghi_poa = project_poa[['poa_global']]
                location = self.location
                daylight_params = DaylightParams(location.latitude, location.longitude, location.tz)
                proj_ghi_bool = ~det.daylight_zeroes(df=proj_ghi_poa, project_params=daylight_params)

                # Convert all False values (bad data) to np.inf so we can replace with NaN after reindexing in the next step
                proj_ghi_bool = proj_ghi_bool.replace(False, np.inf)

                # Reindex to add nighttime rows back in & set them to False so they don't change the data in the original df
                proj_ghi_bool = proj_ghi_bool.reindex(index=proj_ghi_poa.index).fillna(False)

                # Replace np.inf with np.nan (we used np.inf to demarcate bad data so it wouldn't be overwritten in the step above)
                proj_ghi_bool = proj_ghi_bool.replace(np.inf, np.nan)

                # Filter out bad data
                proj_ghi_poa = (proj_ghi_poa*proj_ghi_bool).astype(float)

                # Fill in any NaNs with satellite POA
                proj_ghi_poa = proj_ghi_poa.fillna(df_solcast['sat_poa'])

                # Rename & drop column `sites_ghi_poa` column to match sites with no GHI
                df_solcast['poa'] = proj_ghi_poa
                df_solcast = df_solcast.fillna(0)
            except:
                error_info = sys.exc_info()
                error_type = error_info[:2]
                traceback = error_info[2]
                error_source = traceback.tb_frame.f_code # type: ignore
                lineno = traceback.tb_lineno # type: ignore
                file = error_source.co_filename
                func = error_source.co_name
                
                print(f"Error while fetching 5 min GHI data for {self.name}, using satellite POA instead.")
                print(f"Error details: {error_type} on line {lineno} of `{func}` in {file}")

        
        # Resample high frequency data to hourly
        if resample:
            df_solcast = df_solcast.resample('H').mean()

        #add in a tmod column based on the conversion equation
        a_module = self.array.temperature_model_parameters['a']
        b_module = self.array.temperature_model_parameters['b']
        df_solcast['Tmod'] = df_solcast['Tamb']+(df_solcast['poa']*np.exp(a_module+b_module*df_solcast['Wind_speed']))

        #correct for static values that solcats reports that the dashboard is gonna turn off.
        df_solcast['Correction_factor'] = np.where(df_solcast.index.hour%2==0, 0.999, 1.001) # type: ignore
        df_solcast['Tamb'] *= df_solcast['Correction_factor']
        df_solcast['Tmod'] *= df_solcast['Correction_factor']
        df_solcast['Wind_speed'] *= df_solcast['Correction_factor']

        return df_solcast


# -*- coding: utf-8 -*-
import pandas as pd
from ccrenew.dashboard.data_processing.battery_deg import Battery

#username = os.getenv('username')
#if username not in ['BLUMENTHAL', 'SEBASTIAN', 'Kevin Anderson']:
#    raise RuntimeError("Username not recognized: '%s' -- exiting" % username)
#
#file_project = os.path.join(r'C:\Users', username, r'Box Sync\Cypress Creek Renewables\Asset Management\8) Production Data\_Dashboard_Project')

def Run_AC_Batt(df_kwh, rates, deg_hours):
    #rates = df_Pvsyst_2['Rates'] or df rates
    #df_kwh = df_Pvsyst_2['Year 0 Actual Production (kWh)'] series to time shift, energy at switch gear
    
    #sys.path.append(r'C:\Users\BLUMENTHAL\Documents\Bitbucket\Dispatcher\dispatcher')
    # initiate 
    soc = 1065.864
    b = Battery()
    b.power = 499
    b.nameplate_duration = soc/499
    b.limit = 499
    b.charge_eff = 0.88
    b.discharge_eff = 1.0
    b.degradation_hours = deg_hours
    
    # run
    test = b.binary(rates.values, df_kwh.values, soc0 = soc)
    test['SOC'] = test['SOC'][:-1]
    test['inpS'] = df_kwh
    
    df_battery = pd.DataFrame(test, index = df_kwh.index)

    return df_battery



def AC_Batt_PP(df, df_battery):
    POI_limit = 499
    site_limit = 500
    no_load = 0.5 
    copper_loss = site_limit * 0.009
    
    aux = pd.DataFrame([], index = df.index)
    aux['solar+storage'] = df_battery['solar+storage']
    
    #
    aux['MF_XFMR'] =  aux['solar+storage'] - (no_load + (aux['solar+storage']/site_limit)**2 * copper_loss)
    
    #
    AC_wiring_loss = 0.006
    XFMR_year_max = site_limit * 489.646/500.
    aux['AC_wiring'] = aux['MF_XFMR'] * (1-AC_wiring_loss*(aux['MF_XFMR']/XFMR_year_max)**2)
    
    #skipping HV XFMR
    #skipping gen tie
    #skipping PF loss
    #skipping tracker losses
    
    #
    inv_night_loss = 0.570214669
    aux['Inv_night'] = aux['AC_wiring'].copy()
    aux['Inv_night'].loc[aux['Inv_night']<0] = aux['Inv_night'] - inv_night_loss
    
    #
    scada_loss = 1000
    aux['Scada'] = aux['Inv_night'] - (scada_loss*site_limit/8760./1000.)

    ava = 0.99
    aux['POI_cap'] = aux['Scada'].clip(upper=POI_limit)
    aux['ava'] = aux['POI_cap']*ava
    aux['POI'] = aux['ava'].copy()
    aux['POI_no_night'] = aux['POI'].clip(lower=0)

    
    return aux

#if __name__ == '__main__':
#
#    from CCR import df_keys, file_project
#    import sys
#    
#    sys.path.append(os.path.join(file_project, 'Python_Functions'))
#    from Rate_Structure_Python_with_DST_v07 import generate_Rate_column
#    
#    project_list = df_keys.Project[df_keys.Fund == 'Brunswick'].tolist()
#    
#    # generate rates once, use for all projects
#    file_config = r'C:\Users\Kevin Anderson\Box Sync\Cypress Creek Renewables\Asset Management\8) Production Data\_Dashboard_Project\Projects\Hook 52\Plant_Config_File\Hook 52_Plant_Config_MACRO.xlsm'
#    rates, normal_rate_flag = generate_Rate_column(file_config, shift_DST= True, year = 2015)
#    
#    lis = []
#    for project_name in project_list:
#        
#        # read 8760
#        project_directory = df_keys.Folder[df_keys.Project == project_name].item()
#        if pd.isnull(project_directory):
#            continue
#        file_config = os.path.join(file_project, project_directory, project_name, 'Plant_Config_File', project_name + "_Plant_Config_MACRO.xlsm")
#        df_8760 = pd.read_excel(file_config, sheetname = '8760')
#    
#        # run battery models
#        df_battery = Run_AC_Batt(df_8760['Year 0 Actual Production (kWh)'], rates['Rates'][:-1])
#        
#        df_battery_pp = AC_Batt_PP(df_8760, df_battery)
#        
#        aux = pd.concat([df_8760['POI Output (kWh)'], 
#                         df_battery_pp['POI_no_night'], 
#                         df_8760['Year 0 Actual Production (kWh)'].reset_index(drop=True), 
#                         rates['Rates'][:-1].reset_index(drop=True)
#                        ], axis=1)
#        aux.index = pd.date_range('2015-01-01', '2016-01-01', freq = 'h')[:-1]
#        #aux.plot()
#        lis.append({'Project' : project_name, 
#                    '8760 kWh' : aux['POI Output (kWh)'].sum(), 
#                    'Python kWh' : aux['POI_no_night'].sum(),
#                    'Python rev' : (aux['POI_no_night'] * aux['Rates']).sum(),
#                    '8760 rev' : (aux['POI Output (kWh)'] * aux['Rates']).sum()
#                    })
#        
#        
#    df = pd.DataFrame(lis)
#    df['diff kWh'] = df['8760 kWh'] / df['Python kWh'] -1
#    df['diff rev'] = df['8760 rev'] / df['Python rev'] -1
#    print df
#    
#    '''
#    # old code
#    
#    df_battery = Run_AC_Batt(df['Meter_Corrected_2'], df['Rates'])
#    aux = AC_Batt_PP(df, df_battery)
#    df_battery['POI'] = aux['POI']
#    df_battery.plot()
#    aux2 = pd.concat([df[['POI_Corrected_2', 'Meter_Corrected_2']], df_battery['solar+storage'], aux['POI']], axis=1)
#    aux2['batt discharge-charge'] = df_battery['discharge'] - df_battery['charge'] 
#    aux2.plot()
#    '''


# -*- coding: utf-8 -*-
"""
Created on Thu Jul 28 16:08:53 2016

@author: SEBASTIAN

LOG:

1) 2016-09-25  Inverter Cum Meter.  See Boseman, month of August.  
        Meter gets corrected with Inverter data, but needs cum meter
"""

import pandas as pd
import numpy as np


def calculate_inverter_availability_and_loss(df , df_Pvsyst, P_exp): 
    #
    table_P = pd.pivot_table(df_Pvsyst, values = 'Year 0 Actual Production (kWh)', index =['Month'], columns ='Hour', aggfunc = np.mean)

    #  Find position of Inverter data        
    pos_Inv = [s for s in df.columns if 'Inv_kw_' in s]
    n_inv = len(pos_Inv)
    #    
    pos_Meter = [s for s in df.columns if 'Meter_kw_' in s]
    pos_Meter_cum = [s for s in df.columns if 'Meter_kwnet' in s]
    #
    ava = pd.DataFrame()
    ava[pos_Meter + pos_Meter_cum + pos_Inv] = df[pos_Meter + pos_Meter_cum + pos_Inv]
    #  Inveter cum meter   
    
    pos_Inv_cum = [s for s in df.columns if 'Inv_kwnet' in s]
    ava[pos_Inv_cum] = df[pos_Inv_cum]
    
    #  Invertes is ON if min DC power
    clipping_KW  = df_Pvsyst['Year 0 Actual Production (kWh)'].max()
    clipping_inv = clipping_KW / n_inv
    #inv_on = clipping_inv * .05 # KW
    inv_on = np.min([25, 0.05 * clipping_inv])   #added np.min function for string inverters issue 
    
    #added by saurabh for OM_Uptime on 5/23/18    
    inv_on_OM = 1 #use value of 1 as threshold to determine if inv is on
    ava['Inv_ON_OM'] = ava[ava[pos_Inv] > inv_on_OM].count(axis =1)
    
    #
    #        
    ava['Inv_sum'] = ava[pos_Inv].sum(axis = 1).values 
    #
    ava['Inv_ON'] = ava[ava[pos_Inv] > inv_on].count(axis =1)
    #   
    ava['Inv_avg'] = ava[ava[pos_Inv] >0 ].mean(axis = 1).values
    #
    #  Call Weather sensors to identify DAS ON
    #
    ava['POA_avg'] = df['POA_avg'].values
    ava['Tmod_avg'] = df['Tmod_avg'].values
    ava['Tamb_avg'] = df['Tamb_avg'].values
    ava['Wind_speed_avg'] = df['Wind_speed'].values
    #    
    ava['Meter'] = df['Meter'].values
    ava['Meter_cum'] = df['Meter_cum'].values
    #
    #
    '''  **********************************************************************
    ****     Change 2016-09-25.  Correct Meter_cum based on Meter 
    Correct data outages that last only one day and Inverter Meter has increased'''
    #
    aux = pd.DataFrame()
    aux[pos_Meter + pos_Meter_cum + pos_Inv + pos_Inv_cum] = ava [pos_Meter + pos_Meter_cum + pos_Inv + pos_Inv_cum ]
    aux['Inv_Cum'] = aux[pos_Inv_cum].sum(axis=1)
    aux['Meter_sum'] = ava['Meter']

    aux['Inv_CUM_OFF'] = aux[pos_Inv + pos_Inv_cum].sum(axis = 1).values
    aux['Inv_CUM_OFF_2'] = 0
    aux.loc[aux['Inv_CUM_OFF'] == 0, 'Inv_CUM_OFF_2'] = 1
    aux_i = aux[aux['Inv_CUM_OFF_2'] == 1]
    aux['Inv_CUM_OFF_3'] = 0
    #
    for j in range (0, len(aux_i)):
        aux.loc[aux.index == aux_i.index[j] ,'Inv_CUM_OFF_3'] = 1
        aux.loc[aux.index == aux_i.index[j] +  pd.DateOffset(hours = 1),'Inv_CUM_OFF_3'] = 1
        aux.loc[aux.index == aux_i.index[j] + pd.DateOffset(hours = -1) ,'Inv_CUM_OFF_3'] = 1
    # find increments in cum meter
    aux['Inv_CUM_OFF_4']  = 0
    aux.loc[(aux['Inv_CUM_OFF_2']== 0) & (aux['Inv_CUM_OFF_3'] == 1), 'Inv_CUM_OFF_4']  =1  
    
    # filter out 1-hour data rows in an outage
    # this prevents the start of one outage being the same as the end of the previous, which confuses the script.
    # if 'Inv_CUM_OFF_2' in this row is 0, and the previous row is 1, and the next row is 1, reject it
    aux_curr = aux['Inv_CUM_OFF_2']
    aux_prev = aux['Inv_CUM_OFF_2'].shift(1)
    aux_next = aux['Inv_CUM_OFF_2'].shift(-1)
    bad_indices = list((aux_curr == 0) & (aux_prev == 1) & (aux_next == 1))
    aux.loc[bad_indices, 'Inv_CUM_OFF_4'] = 0
    
    #
    aux_column_name =  ["Cum_" + c for c in pos_Inv_cum]
    aux_Missing = pd.DataFrame( columns = aux_column_name)
    aux_Missing[aux_column_name] = aux.loc[aux['Inv_Cum'] == 1, pos_Inv_cum].diff()[1::2]
    aux_Missing  = aux_Missing.sum(axis =1 )
    #   
    #aux_2.loc[aux_2[aux_column_name] < 0,aux_column_name] = 0                           #cum meter difference can't be negative    
    #
    #  Inverter Expected Production.  Based on historical data.
    #
    aux_inv_exp = ava.loc[ (ava['Inv_ON'] == n_inv), [ 'Inv_ON','Inv_sum']  ]   #get inv sum where all inverters is on
    #####START BACK HERE CHRIS
    try:
        aux_inv_exp['hour'] = aux_inv_exp.index.hour
    except AttributeError:
        aux_inv_exp = aux_inv_exp.reindex(columns=aux_inv_exp.columns.to_list() + ['hour'])
    
    aux_inv_exp = aux_inv_exp.groupby( aux_inv_exp['hour'].values).mean() #get hourly average of inv sum
        
    aux_inv_exp_2 = pd.DataFrame(index = range(0,24))
    aux_inv_exp = pd.concat([aux_inv_exp_2, aux_inv_exp], axis = 1).fillna(0) #fill in the rest of the 24 hours
    #
    aux_inv_exp = aux_inv_exp['Inv_sum']
    #
    fill_inv = pd.DataFrame()
    fill_inv[aux.columns] = aux[aux['Inv_CUM_OFF_2'] == 1]
    try:
        fill_inv['hour'] = fill_inv.index.hour
    except AttributeError:
        fill_inv = fill_inv.reindex(columns=fill_inv.columns.to_list() + ['hour'])
    #
    aux_fill = []
    date_fill = []
    for i in range(0, len( fill_inv.index)):
        hour_fill = fill_inv.index.hour[i]
        aux_fill.append(aux_inv_exp[aux_inv_exp.index == hour_fill].values[0])
        date_fill.append( fill_inv.index[i])
    #    
    #aux_fill = pd.DataFrame(aux_fill, index = date_fill)                       #changed by ryan due to crashing when no meter outages
    aux_fill = pd.DataFrame(aux_fill, index = date_fill, columns = ['Inv_sum_Fill'])  
    aux_fill = aux_fill.set_index(aux[aux['Inv_CUM_OFF_2'] == 1].index)
    #aux_fill.columns = ['Inv_sum_Fill']                                         #changed by ryan due to crashing when no meter outages
    #
    aux_f = pd.concat([aux, aux_Missing, aux_fill], axis=1)
    aux_f = aux_f.fillna(0)
    aux_f.rename(columns={0:'Missing_Inv'}, inplace=True)
    
    #  Distribute energy
    pos = aux_f.loc[aux_f['Inv_CUM_OFF_4'] > 0,  'Inv_CUM_OFF_4'].index 
    #
    
    # if the file starts out with missing data, then the end of this block will mistakenly be
    # identified as the beginning of a block.  Check to see if this is the case, and correct if necessary
    if aux_f['Inv_CUM_OFF_2'][0] == 1:
        print("*** Data started in an outage")
        pos = pos[1:]
    # similarly, if the file ends with missing data, remove the last element of pos
    if aux_f['Inv_CUM_OFF_2'][-1] == 1:
        print("*** Data ended in an outage")
        pos = pos[:-1]
    
    if (len(pos) % 2) == 1: # "odd"
        print("*** Number of (start, end) outage pairs is mismatched")
    #     
    
    
    aux_e_filled_value = []
    aux_e_filled_date = []
    for p in range(0,len(pos),2):
        e_to_distribute = aux_f['Inv_sum_Fill'][pos[p]:pos[p+1]] - aux_f['Meter_sum'][pos[p]:pos[p+1]]
        e_to_distribute = e_to_distribute.sort_values( ascending = False)
        e_to_fill = aux_f['Missing_Inv'][pos[p+1]]    
        #
        k = 0
        while k < len(e_to_distribute):
            if e_to_distribute[k] < e_to_fill:
                aux_e_filled_value.append(e_to_distribute[k])
                aux_e_filled_date.append(e_to_distribute.index[k])
                e_to_fill = e_to_fill - e_to_distribute[k]                
            k = k+1
        #        
    aux_e_filled = pd.DataFrame(aux_e_filled_value, index = aux_e_filled_date)
    aux_f2 = pd.concat([aux, aux_f[['Missing_Inv' ,'Inv_sum_Fill']], aux_e_filled], axis=1) 
    if aux_e_filled.empty:
        aux_f2['Missing_Inv_ADDED'] = 0
    else:            
        
        aux_f2 = aux_f2.fillna(0)
        aux_f2.rename(columns={0:'Missing_Inv_ADDED'}, inplace=True)    
        aux_f2['Missing_Inv_ADDED'] = aux_f2['Missing_Inv_ADDED'].multiply(aux_f2['Inv_CUM_OFF_2'])
    #
    #  Correct Meter, with Column named 0
    ava['Meter_ORIGINAL'] = ava['Meter']
    #ava['Meter'] = ava['Meter'] + aux_f2['Missing_Inv_ADDED'].multiply(aux_f2['Inv_CUM_OFF_2'])    #do not perform correction anymore 10/19
    #
    
    #   TO troubleshoot use  aux_f2.to_clipboard()   
    
    #
    # END CHANGE 1
    '''  ******************************************************************'''
    #
    #--------------------------------------------------------------------------
    #  DAS is considered ON if values are reported for at least one of the following POA, Meter and INV
    ava['DAS_ON'] = df['DAS_ON']
    #    ava['DAS_ON'] = (ava).sum(axis=1).values
    #    ava.loc[ava['DAS_ON'] >0, 'DAS_ON' ] = 1  
    #    #  Remove Temperature outliers
    #    ava.loc[ava['DAS_ON'] <0, ['DAS_ON', 'Tmod_avg', 'Tamb_avg'] ] = 0  
    #
    #Find how many inverters are ON based on Meter production
    ava['N_inv_prod'] = ava['Meter'].div(ava['Inv_avg'] ).fillna(0).round().values #sum of meters / inv avg (rounded)
    ava.loc[ava['N_inv_prod']  > n_inv , ['N_inv_prod']] =  n_inv #
    #
    ava['N_inv_prod'] = ava[['N_inv_prod', 'Inv_ON']].max(axis =1).values
    #
    ava['AVA'] = ava['N_inv_prod'].div(n_inv).values
    # fix times where one inverter is off, but the others are picking up the slack
    ava_min = (ava['Meter'] / (clipping_KW*1)).clip(upper=1)
    ava.loc[n_inv * ava['Inv_avg'] >= (clipping_KW*1), 'AVA'] = ava['AVA'].clip(lower=ava_min)

    
    #ADDED BY SAURABH for OM Uptime on 5/25/2018
    #find out #inverters on using met production
    ava['N_inv_prod_OM'] = ava['Meter'].div(ava['Inv_avg'] ).fillna(0).round().values 
    ava.loc[ava['N_inv_prod_OM']  > n_inv , ['N_inv_prod_OM']] =  n_inv 
    ava['N_inv_prod_OM'] = ava[['N_inv_prod_OM', 'Inv_ON_OM']].max(axis =1).values 
    
    #calc om uptime
    ava['OM_Uptime'] = ava['N_inv_prod_OM'].div(n_inv).values
    #
    
     #Fix values where AVA > 0 but less than 1  
    ava['Meter_&_ava'] = ava['Meter'].divide(ava['AVA']) #.multiply((len(pos_Inv) * ava['AVA'])/len(pos_Inv))
    
    #Fix values where AVA = 0
    condition = ava.loc[ (ava['AVA'] == 0) & (df['POA_avg'] > 100) & (ava['DAS_ON'] ==1), 'AVA'].index
    #condition = ava.loc[ (ava['AVA'] == 0) & (df['POA_avg'] > 100), 'AVA'].index

    ava.loc[condition, 'Meter_&_ava'] = P_exp
    
    #Make corrections
    ava.loc[ ava['Meter_&_ava'] == np.inf, 'Meter_&_ava'] = ava['Meter']
    ava.loc[ ava['Meter_&_ava'] == -np.inf, 'Meter_&_ava'] = ava['Meter']
    ava['AVA_Energy_loss'] = ava['Meter_&_ava'] - ava['Meter']
    
    #ava['AVA_Energy_loss'][ava['AVA_Energy_loss'] <0 ] = 0
    ava.loc[(ava['AVA_Energy_loss'] <0) , ['AVA_Energy_loss']] = 0
    ava.loc[(ava['Inv_ON'] <= 0) , ['AVA_Energy_loss']] = 0
    #
    #--------------------------------------------------------------------------
    #    CORRECT  METER DATA
    #--------------------------------------------------------------------------
    #  find rows with INV and Weather data but Not Meter
    #
    #TODO: df_sensor_ON[pos_Inv].all(axis=1)
    n = ava[(ava['POA_avg'] >= 100) & (ava['Inv_ON'] == n_inv) & (ava['Meter'] == 0)]
    n['Meter_Corrected'] = n[pos_Inv].sum(axis = 1) 
       
    ava['Meter_Corrected'] = n['Meter_Corrected']
    ava = ava.fillna(0)
    ava['Meter_Corrected'] = ava[['Meter_Corrected', 'Meter']].sum(axis = 1) #this works because Meter is 0 where meter corrected is made
    #
    #  Find rows when all data is missing
    ava['ALL_OFF'] = 0 # FALSE
    ava.loc[ava['DAS_ON'] == 0, 'ALL_OFF'] = 1  # TRUE
    #ava.loc[ava.sum(axis =1) ==0, 'ALL_OFF'] = 1  # TRUE
    #
    
    
    #correct meter using inverter sum
        #  Rows with All_Off = 0, check if the cum meter is ok, correct with Meter_Corrected
        #aux = ava[(ava['ALL_OFF'] != 1) & (ava['Meter_cum'] <=0 ) & (ava['Meter_Corrected'].diff(1) != 0),'Meter_Cum_Corrected'] = ['Meter_cum'].
    ava['Meter_cum_corrected'] = ava['Meter_cum']
    index = list(range(len(ava)))                        #define i = 1 to end of array
    index.extend(list(reversed(range(len(ava)))))   #extend array to include reverse direction (required for correcting upwards)

    for i in index:
        #check if all off is true, cum meter is missing and the next metered value exists. If so, subtract (go upwards in excel)
        if (i < (len(ava)-1)):
            if ((ava['DAS_ON'].values[i] == 1) & (ava['Meter_cum_corrected'].values[i] <=0) & (ava['Meter_cum_corrected'].values[i+1] > 0)):
                ava['Meter_cum_corrected'].values[i] = ava['Meter_cum_corrected'].values[i+1] - ava['Meter_Corrected'].values[i]
            
        #check if all off is true, cum meter is missing and the previous metered value exists. If so, add (go downards in excel)
        if(i > 0):
            if ((ava['DAS_ON'].values[i] == 1) & (ava['Meter_cum_corrected'].values[i] <=0) & (ava['Meter_cum_corrected'].values[i-1] > 0)):
                ava['Meter_cum_corrected'].values[i] = ava['Meter_cum_corrected'].values[i-1] + ava['Meter_Corrected'].values[i]
            
    #--------------------------------------------------------------------------
    #
    #  Find ALL_OFF =1 and   Find start and end of Grid outages
    aux = ava['ALL_OFF'] +ava['ALL_OFF'].shift(-1).fillna(0)+ ava['ALL_OFF'].shift(1).fillna(0) 
    aux.loc[aux>0 , ] =1
    ava['ALL_OFF_2'] = aux
    ava['ALL_OFF_3'] = ava['ALL_OFF_2'] -ava['ALL_OFF']
    #
    #  Calculate energy produced in Intervals ALL_OFF_3
    ava['ALL_OFF_3_P'] = 0
    aux = ava.loc[ava['ALL_OFF_3'] == 1, 'Meter_cum_corrected']     #check for large jumps in cum meter when difference = 0
    if (len(aux) % 2) == 1: # "remove odd outages - last outage doesn't matter"
        aux = aux[:-1] 
    
    for i in range(0,len(aux),2):       #if cum meter difference involves 0, then ignore difference
        if (aux[i] * aux[i+1] == 0):
            aux[i]= 0
            aux[i+1]=0
   
    #ava['Meter_cum_corrected'][aux.index] = aux.values
    ava.loc[aux.index, 'Meter_cum_corrected'] = aux.values      #substitute data back in
    
    # filter out 1-hour data rows in an outage
    # this prevents the start of one outage being the same as the end of the previous, which confuses the script.
    # if 'ALL_OFF_2' in this row is 0, and the previous row is 1, and the next row is 1, reject it
    ava_curr = ava['ALL_OFF']
    ava_prev = ava['ALL_OFF'].shift(1)
    ava_next = ava['ALL_OFF'].shift(-1)
    bad_indices = list((ava_curr == 0) & (ava_prev == 1) & (ava_next == 1))

    ava.loc[bad_indices, 'ALL_OFF_3'] = 0
    
    ava['ALL_OFF_3_P'] = ava.loc[ava['ALL_OFF_3'] == 1, 'Meter_cum_corrected'].diff()[1::2]
    ava.loc[ava['ALL_OFF_3_P'] < 0,'ALL_OFF_3_P'] = 0                           #cum meter difference can't be negative
    
    #
    #
    # For Every Hour with ALL OFF calculate "Expected_PVsyst generation"      
    m_h_list = ava.loc[ava['ALL_OFF'] == 1, ].index
    #
    ava['PVsyst_Expect'] = 0
    for m_h in m_h_list :
        hour = m_h.hour
        month = m_h.month
        p_exp = table_P.loc[(table_P.index == month)][hour]
        ava.loc[ava['PVsyst_Expect'].index ==  m_h, 'PVsyst_Expect' ] = p_exp.values[0]     
    
    
    
    #
    #  Find rows wit ALL_OFF_3_P
    # 
    aux = ava.loc[ava['ALL_OFF_2']>0,  ['ALL_OFF_2','ALL_OFF_3']] 
    pos = ava.loc[ava['ALL_OFF_3']>0,  'ALL_OFF_3'].index 
    #
    
    ava['ALL_OFF_3_P_org'] = 0
       
    # if the file starts out with missing data, then the end of this block will mistakenly be
    # identified as the beginning of a block.  Check to see if this is the case, and correct if necessary
    if ava['ALL_OFF'][0] == 1:
        print("*** Data started in an outage")
        pos = pos[1:]
    # similarly, if the file ends with missing data, remove the last element of pos
    if ava['ALL_OFF'][-1] == 1:
        print("*** Data ended in an outage")
        pos = pos[:-1]
    
    if (len(pos) % 2) == 1: # "odd"
        print("*** Number of (start, end) outage pairs is mismatched")

    for p in range(0,len(pos),2):
        e_distribute = ava.loc[ava.index == pos[p+1] , ['ALL_OFF_3_P']].fillna(0).values[0][0]
        e_pvsyst = ava.loc[(ava.index >= pos[p]) & (ava.index <= pos[p+1]) , ['PVsyst_Expect']]
        #  Sort values
        e_pvsyst = e_pvsyst.sort_values(by ='PVsyst_Expect' ,ascending = False)  
        e_pvsyst = e_pvsyst.loc[ava['ALL_OFF'] == 1] 
        #
        if len(e_pvsyst) == 1:
            ava.loc[ava['ALL_OFF_3_P_org'].index == e_pvsyst.index[0] ,'ALL_OFF_3_P_org'] = e_distribute
        else:
            k = 0  
            #  Loop all hours and spread the Energy produced.
            while k < len(e_pvsyst) :
                if e_distribute <= e_pvsyst['PVsyst_Expect'][k]:
                    ava.loc[ava['ALL_OFF_3_P_org'].index == e_pvsyst.index[k] ,'ALL_OFF_3_P_org'] = e_distribute
                    k = len(e_pvsyst)                    
                else:
                     ava.loc[ava['ALL_OFF_3_P_org'].index == e_pvsyst.index[k] ,'ALL_OFF_3_P_org'] = e_pvsyst['PVsyst_Expect'][k] 
                     e_distribute = e_distribute  - e_pvsyst['PVsyst_Expect'][k]               
                k = k+1
                #  Check if there is energy left at e_distribute
                if (k == len(e_pvsyst)) & (e_distribute > 0) :
                        ava.loc[ava['ALL_OFF_3_P_org'].index == e_pvsyst.index[k-1] ,'ALL_OFF_3_P_org'] = e_distribute + e_pvsyst['PVsyst_Expect'][k-1] 
    #
    #                  I have no idea what all_off-3_p_org means- Chris
    ava['Meter_Corrected_2'] =   ava['Meter_Corrected'] +  ava['ALL_OFF_3_P_org']   
    #
    ava['Grid_loss'] = ava['PVsyst_Expect'] - ava['ALL_OFF_3_P_org']
    ava.loc[(ava['Grid_loss'] < 0), 'Grid_loss'] = 0   
    #
    ava['GRID_AVA'] = 1  
    #
    aux = ava['ALL_OFF_3_P_org'].div(ava['PVsyst_Expect']).fillna(1)
    aux.loc[aux >1] = 1
    #
    #ava.loc[ava['ALL_OFF'] == 1, 'GRID_AVA'] = 0#(ava['ALL_OFF_3_P_org'].values / ava['PVsyst_Expect'].values )
    ava['GRID_AVA'] = aux
    
    ava['Meter_cum_corrected_2'] = ava['Meter_Corrected_2'].cumsum() #+ ava['Meter_cum_corrected'][1] #-ava['Meter_Corrected_2'][1]
    
    ava.loc[(ava['Meter_Corrected_2'] > ava['Meter_&_ava']), 'Meter_&_ava'] = ava['Meter_Corrected_2']
    #ava.to_clipboard()        
    return ava


#!/usr/bin/env python
from ccrenew import __version__
from setuptools import setup

setup(
    name='ccrenew',
    version=__version__,
    author='CCR Perf Eng Team',
    description='Utilities for running Cypress Creek Renewables Performance Engineering workflows',
    url='https://cypresscreekrenew.sharepoint.com/:u:/s/AM-Performance/EdhVQHbS4n9Jm022irFSFiABmo6gcEW3KTmV0qEkfHW84Q?e=zxokuU',
    python_requires='>=3.8'
)


import awswrangler as wr
import boto3
from dateutil import parser
import pandas as pd
from pathlib import Path
from multiprocessing.pool import ThreadPool
import s3fs

from ccrenew import (
    all_df_keys as df_keys,
    ccr,
    DateLike,
    Numeric,
    ProjectModelParams,
    AWSError,
)

s3_config = boto3.client('s3')
bucket = 'perfdatadev.ccrenew.com'
bucket_prefix_satellite = "5min_archive/Solcats/"
bucket_prefix_measured = "5min_archive/PF/"


# https://preview-docs.solcast.com.au/


def get_sat_weather(project_name:str, start:str|DateLike, end:str|DateLike,
                    convert:bool = False, pool_size:int = 6):
    """
    Query data from AWS to use in irradiance calculations
    
    Args:
        ccr_id (str): CCR ID for the project.
        start (str or DateLike): Start date for the data request.
        end (str or DateLike): Start date for the data request.
        source (str): Option to pull data from satellite recordings or on-site
            measurements. Options: `satellite` or `measured`
        pool_size (int): Number of threads to use when pulling data directly from S3.
    
    Returns
        df_weather (pd.DataFrame): Weather for the project over the specified days.
    """
    ccr_id, folder = _get_params(project_name)

    keys, (start, end) = _generate_keys(ccr_id, start, end, 'satellite')
    if len(keys) < 50:
        df_weather = _s3_pool(pool_size=pool_size, keys=keys)
        df_weather = df_weather.reindex(columns=['ghi', 'dni', 'dhi', 'air_temp', 'wind_speed_10m'])
    else:
        df_weather = wr.athena.read_sql_query(
            sql="""  SELECT  local_datetime
                            ,ghi
                            ,dni
                            ,dhi
                            ,air_temp
                            ,wind_speed_10m
                       FROM raw_data
                      WHERE ccr_id=:ccr_id;
                        AND local_datetime>=:start;
                        AND local_datetime<=:end;
                   ORDER BY local_datetime""",
            params={"ccr_id":f"'{ccr_id}'",
                    "start":f"CAST('{start.strftime('%Y-%m-%d')}' AS DATE)",
                    "end":f"CAST('{end.strftime('%Y-%m-%d')}' AS DATE)"
                    },
            database='solcast',
            ctas_approach=False
            )
        df_weather = df_weather.set_index('local_datetime') # type: ignore

    if df_weather.empty:
        raise AWSError(f"Error querying weather data from AWS")
    else:
        df_weather = df_weather.rename(columns={'wind_speed_10m': 'Wind_speed',
                                                'air_temp': 'Tamb',
                                                'ghi': 'sat_ghi'})

        if convert:
            # Convert units
            folder = folder
            config_file = project_name + r'_Plant_Config_MACRO.xlsm'
            config_path = Path(ccr.file_project) / folder / project_name / 'Plant_Config_File' / config_file
            xl=pd.ExcelFile(config_path)
            df_config=xl.parse('Unit_Tab')
            
            #temp units
            temps = df_config.loc[df_config['Var_ID'].str.contains('amb')]['Convert_Farenheit_to_Celcius'].sum()
            temps += df_config.loc[df_config['Var_ID'].str.contains('mod')]['Convert_Farenheit_to_Celcius'].sum()
            if temps > 0:
                df_weather['Tamb'] = df_weather['Tamb']*float((9/5.))+32.0
                df_weather['Tmod'] = df_weather['Tmod']*float((9/5.))+32.0
            else:
                df_weather['Tamb'] = df_weather['air_temp']
                
            #wind units
            winds = df_config.loc[df_config['Var_ID'].str.contains('speed')]['Convert_mph_to_mps'].sum()
            if winds > 0:
                df_weather['Wind_speed'] = df_weather['wind_speed_10m']*float(2.23694)
            else:
                df_weather['Wind_speed'] = df_weather['wind_speed_10m']

    return df_weather

def get_project_ghi(project_name:str, start:str|DateLike, end:str|DateLike,
                    ghi_index:int = 0, pool_size:int = 6):
    ccr_id, _ = _get_params(project_name)
    keys, (start, end) = _generate_keys(ccr_id, start, end, 'project_ghi')
    df_project_ghi = _s3_pool(pool_size=pool_size, keys=keys)
    ghi_col = [col for col in df_project_ghi.columns if 'GHI' in col][ghi_index]
    df_project_ghi = df_project_ghi[[ghi_col]].rename(columns={ghi_col: 'proj_ghi'})

    return df_project_ghi

def _make_satellite_key(CCR_ID, date):
    return bucket_prefix_satellite + CCR_ID + "/" + "sat_weather_" + date.strftime('%Y-%m-%d') + ".csv"

def _make_project_key(CCR_ID, date):
    return bucket_prefix_measured + CCR_ID + "/" + "main_" + date.strftime('%Y-%m-%d') + ".csv"

def _retrieve_df(key):
    fs = s3fs.S3FileSystem(anon=False)
    path = "s3://{b}/{k}".format(b=bucket, k=key)
    try:
        with fs.open(path, 'rb') as f:
            df = pd.read_csv(f, index_col = 0, parse_dates = True)
            df = df.loc[~df.index.duplicated(), :]
    except FileNotFoundError:
        df = pd.DataFrame()
    return df

def _get_project_model_params(project_name:str) -> ProjectModelParams:

    df_proj_keys = df_keys.query("Project == @project_name").to_dict('records')[0]

    project_params = ProjectModelParams(
        project_name,
        df_proj_keys['CCR_ID'],
        df_proj_keys['Folder'],
        f"US/{df_proj_keys['Timezone']}",
        df_proj_keys['GPS_Lat'],
        df_proj_keys['GPS_Long'],
        df_proj_keys['Elevation'],
        df_proj_keys['MWDC'],
        df_proj_keys['Racking'],
        df_proj_keys['Tilt/GCR'],
        df_proj_keys['Max_angle'],
        df_proj_keys['Temp_Coeff_Pmax'],
        df_proj_keys['a_module'],
        df_proj_keys['b_module'],
        df_proj_keys['Delta_Tcnd'],
    )

    return project_params

def _get_params(project_name:str):
    project_params = df_keys.query("index == @project_name").to_dict(orient='records')[0]
    ccr_id = project_params['CCR_ID']
    folder = project_params['Folder']

    return ccr_id, folder

def _generate_keys(ccr_id:str, start:str|DateLike, end:str|DateLike, type: str):
    if isinstance(start, str):
        start = parser.parse(start)
    if isinstance(end, str):
        end = parser.parse(end)
    date_range = pd.date_range(start, end, normalize=True)

    if type == 'satellite':
        keys = [_make_satellite_key(ccr_id, date) for date in date_range]
    else:
        keys = [_make_project_key(ccr_id, date) for date in date_range]

    return keys, (start, end)

def _s3_pool(pool_size, keys):
    pool_size = 6
    pool = ThreadPool(pool_size)
    day_df_list = pool.map(_retrieve_df, keys)
    pool.close()
    pool.join()
    
    df_sat_weather = pd.concat(day_df_list, axis=0)
    return df_sat_weather

